﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  [Item("Regression Transformation Model", "A regression model that was transformed back to match the original variables after the training was performed on transformed variables.")]
  [StorableClass]
  public class RegressionTransformationModel : DataAnalysisTransformationModel, IRegressionTransformationModel {

    public new IRegressionModel OriginalModel {
      get { return (IRegressionModel)base.OriginalModel; }
    }

    IEnumerable<IDataAnalysisTransformation> IRegressionTransformationModel.TargetTransformations {
      get { return TargetTransformations; }
    }

    #region Constructor, Cloning & Persistence
    public RegressionTransformationModel(IRegressionModel originalModel, IEnumerable<IDataAnalysisTransformation> transformations)
      : base(originalModel, transformations) {
      var learnedTarget = DataAnalysisTransformation.GetStrictTransitiveVariables(originalModel.TargetVariable, transformations, true).Last();
      var transitiveTargets = DataAnalysisTransformation.GetStrictTransitiveVariables(learnedTarget, transformations, inverse: false).ToList();
      TargetTransformations = new ItemList<IDataAnalysisTransformation>(transformations.Where(t => transitiveTargets.Contains(t.TransformedVariable))).AsReadOnly();

      TargetVariable = DataAnalysisTransformation.GetStrictTransitiveVariables(originalModel.TargetVariable, TargetTransformations, true).Last();
    }

    protected RegressionTransformationModel(RegressionTransformationModel original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new RegressionTransformationModel(this, cloner);
    }

    [StorableConstructor]
    protected RegressionTransformationModel(bool deserializing)
      : base(deserializing) { }
    #endregion

    public virtual IEnumerable<double> GetEstimatedValues(IDataset dataset, IEnumerable<int> rows) {
      var transformedInput = DataAnalysisTransformation.Transform(dataset, InputTransformations);
      var estimates = OriginalModel.GetEstimatedValues(transformedInput, rows);
      return InverseTransform(estimates, TargetTransformations);
    }

    public virtual IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
      return new RegressionSolution(this, new RegressionProblemData(problemData));
    }

    protected static IEnumerable<double> InverseTransform(IEnumerable<double> data, IEnumerable<IDataAnalysisTransformation> transformations) {
      foreach (var transformation in transformations.Reverse()) { // TargetTransformations only contains only relevant transformations
        var trans = (ITransformation<double>)transformation.Transformation;
        data = trans.InverseApply(data).ToList();
      }
      return data;
    }
  }
}